class ImageprobeError(Exception):
    """Generic base error"""


class CorruptedImage(ImageprobeError):
    """Parsing error"""


class DownloadError(ImageprobeError):
    """Download error

    Attributes:
        url (str): Image URL.
        bytes_read (str): Nr. of bytes received up to the exception raise.
    """

    def __init__(self, url: str, bytes_read: int) -> None:
        self.url = url
        self.bytes_read = bytes_read
        message = f'url: "{url}", {bytes_read} bytes read'
        super().__init__(message)


class UnsupportedFormat(ImageprobeError):
    """Filetype is not supported."""



# pylint: disable=useless-import-alias

import asyncio
import sys
from typing import Tuple

# Shortcuts
from .parser import probe as probe

__all__: Tuple[str, ...] = ("probe",)

# Workaround for known issue of aiohttp on Windows / Python 3.8
# See: https://github.com/aio-libs/aiohttp/issues/4324
if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform == "win32":
    # pylint: disable=no-member
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())



import struct
from typing import Optional

from imageprobe.client import DownloadClient
from imageprobe.errors import CorruptedImage
from imageprobe.types import ImageData


async def png(client: DownloadClient) -> Optional[ImageData]:
    # Validate PNG signature
    await client.read_from_start(8)
    if not client.buffer.startswith(b"\211PNG\r\n\032\n"):
        return None

    # Check that IHDR chunk exists
    await client.read_from_start(16)
    if client.buffer[12:16] != b"IHDR":
        raise CorruptedImage("Invalid PNG file: missing IHDR chunk")

    # Load IHDR chunk and unpack image metadata
    await client.read_from_start(24)
    try:
        width, height = struct.unpack(">LL", client.buffer[16:24])
    except struct.error as exc:
        raise CorruptedImage("Invalid PNG file: corrupted IHDR chunk") from exc

    return ImageData(
        width=width,
        height=height,
        extension="png",
        mime_type="image/png",
    )



from typing import Optional

import aiohttp

from imageprobe.client import DownloadClient
from imageprobe.errors import UnsupportedFormat
from imageprobe.parsers import ordered_parsers as parsers
from imageprobe.types import ImageData


async def probe(url: str, session: Optional[aiohttp.ClientSession] = None) -> ImageData:
    """Download as little data as possible to determine the dimensions of an image.

    Args:
        url (str): Image URL.
        session (Optional[aiohttp.ClientSession]): Pre-existing session. Defaults to
            None if you want to instantiate a new one.

    Raises:
        CorruptedImage: Image is corrupted.
        DownloadError: Image wasn't downloaded successfully.
        UnsupportedFormat: Filetype is not supported.

    Returns:
        ImageData: Object containing image metadata.
    """
    async with DownloadClient(url, session) as client:
        _image_data: Optional[ImageData] = None
        for parser in parsers:
            _image_data = await parser(client)
            if _image_data is not None:
                break

        if _image_data is None:
            raise UnsupportedFormat
        return _image_data



import struct
from typing import Optional

from imageprobe.client import DownloadClient
from imageprobe.errors import CorruptedImage
from imageprobe.types import ImageData


async def gif(client: DownloadClient) -> Optional[ImageData]:
    # Validate GIF signature
    await client.read_from_start(6)
    if client.buffer[:6] not in (b"GIF87a", b"GIF89a"):
        return None

    # Fetch dimensions
    await client.read_from_start(10)
    try:
        width, height = struct.unpack("<HH", client.buffer[6:10])
    except struct.error as exc:
        raise CorruptedImage("Invalid GIF file") from exc

    return ImageData(
        width=width,
        height=height,
        extension="gif",
        mime_type="image/gif",
    )



from dataclasses import dataclass


@dataclass(frozen=True)
class ImageData:
    """Contains image metadata.

    Attributes:
        width (int): Width expressed in `w_units`.
        height (int): Height expressed in `h_units`.
        extension (str): File extension, i.e. "png".
        mime_type (str): Mime type, i.e. "image/png".
        w_units (str): Measurement units for `width`. Defaults to "px".
        h_units (str): Measurement units for `height`. Defaults to "px".
    """

    width: int
    height: int
    extension: str
    mime_type: str
    w_units: str = "px"
    h_units: str = "px"



from types import TracebackType
from typing import Optional, Type

import aiohttp

from imageprobe.errors import DownloadError


class DownloadClient:
    """Wrapper for aiohttp.ClientSession.

    It either creates a new private ClientSession instance or uses a pre-existing one if
    it is provided at inizialization.
    Read methods fail at if EOF is reached before reading enough data.

    Attributes:
        url (str): Image URL.
        buffer (bytes): Bytes downloaded from `url`.
        bytes_read (int): Syntactic sugar for `len(buffer)`.
    """

    def __init__(
        self,
        url: str,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        self.url = url
        self.buffer = b""
        self._cs: aiohttp.ClientSession
        self._external_cs = session
        self._response: aiohttp.ClientResponse

    @property
    def bytes_read(self) -> int:
        return len(self.buffer)

    async def __aenter__(self) -> "DownloadClient":
        # If no client session is provided, instantiate a new one.
        if self._external_cs is None:
            self._cs = aiohttp.ClientSession()
        else:
            self._cs = self._external_cs

        try:
            self._response = await self._cs.get(self.url)
            return self
        except aiohttp.ClientError as exc:
            if self._external_cs is None:
                await self._cs.close()
            raise DownloadError(self.url, self.bytes_read) from exc

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: TracebackType,
    ) -> None:
        await self._release()

    async def _release(self) -> None:
        """Gracefully interrupts the connection.

        It also closes the client session if it wasn't provided externally.
        """
        await self._response.release()
        if self._external_cs is None:
            await self._cs.close()

    async def read(self, nr_bytes: int) -> None:
        """Reads `nr_bytes` bytes."""
        try:
            old_buflen = self.bytes_read
            self.buffer += await self._response.content.read(nr_bytes)
        except aiohttp.ClientError as exc:
            await self._release()
            raise DownloadError(self.url, self.bytes_read) from exc

        # Differently from classic implementations, we require to read all nr_bytes (and
        # not up to nr_bytes).
        if self.bytes_read - old_buflen < nr_bytes:
            await self._release()
            raise DownloadError(self.url, self.bytes_read)

    async def read_from_start(self, nr_bytes: int) -> None:
        """Reads `nr_bytes` from the beginning of the file."""
        bytes_diff = nr_bytes - self.bytes_read
        if bytes_diff:
            await self.read(bytes_diff)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imageprobe', 'imageprobe.parsers']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.3,<4.0.0']

extras_require = \
{':python_version >= "3.6.2" and python_version < "3.7.0"': ['dataclasses>=0.7,<0.8']}

setup_kwargs = {
    'name': 'imageprobe',
    'version': '0.2.0',
    'description': 'Asynchronous image probing library. Fetch only as much as you need!',
    'long_description': '# imageprobe\n\n[![Latest PyPI package version](https://img.shields.io/pypi/v/imageprobe.svg)](https://pypi.org/project/imageprobe)\n[![Supported Python versions](https://img.shields.io/pypi/pyversions/imageprobe.svg)](https://pypi.org/project/imageprobe)\n[![Development status](https://img.shields.io/pypi/status/imageprobe)](https://pypi.org/project/imageprobe)\n[![CI](https://github.com/palt0/imageprobe/workflows/CI/badge.svg)](https://github.com/palt0/imageprobe/actions?query=workflow%3ACI)\n[![Codecov](https://codecov.io/gh/palt0/imageprobe/branch/main/graph/badge.svg?token=DIHQIYQJ91)](https://codecov.io/gh/palt0/imageprobe)\n\nAsynchronous library to get image dimensions by fetching as little data as possible.\n\nIt temporarily supports only GIF, PNG because development is still in a very early stage.\n\n## Usage\n\nTo install this library, run:\n\n    pip install imageprobe\n\nThe `probe()` function returns metadata of an image from an URL, or throws an exception if an error occurred.\n\n```python\nimport asyncio\nfrom imageprobe import probe\n\nloop = asyncio.get_event_loop()\nurl = "https://upload.wikimedia.org/wikipedia/commons/7/70/Example.png"\nimage_data = loop.run_until_complete(probe(url))\nprint(image_data.width, image_data.height)\n\n# 172 178\n```\n\nUnder the hood, `probe()` creates an `aiohttp.ClientSession`, but you can pass a pre-existing session as an optional argument if you prefer.\n\n## Contributing\n\nI won\'t accept pull requests until the first beta release.\n',
    'author': 'Plato',
    'author_email': 'platoo@outlook.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/palt0/imageprobe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)
